""" Save forecasts to the database """
import logging
from typing import List, Optional

from sqlalchemy.orm.session import Session

from nowcasting_datamodel.models import PVSystem, PVSystemSQL
from nowcasting_datamodel.models.forecast import ForecastSQL
from nowcasting_datamodel.save.adjust import add_adjust_to_forecasts
from nowcasting_datamodel.save.update import (
    add_forecast_last_7_days_and_remove_old_data,
    change_forecast_value_to_forecast_last_7_days,
    update_all_forecast_latest,
)

logger = logging.getLogger(__name__)


def save(
    forecasts: List[ForecastSQL],
    session: Session,
    update_national: Optional[bool] = True,
    update_gsp: Optional[bool] = True,
    apply_adjuster: Optional[bool] = True,
):
    """
    Save forecast to database

    1. Add sqlalchemy onjects to database
    2. Saves to 'latest' table aswell

    :param forecasts: list of sql forecasts
    :param session: database session
    :param update_national: Optional (default true), to update the national forecast
    :param update_gsp: Optional (default true), to update all the GSP forecasts
    :param apply_adjuster: Optional (default true), to apply the adjuster
    """
    if apply_adjuster:
        logger.debug("Add Adjust to forecasts")
        add_adjust_to_forecasts(session=session, forecasts_sql=forecasts)

    # save objects to database
    logger.debug("Saving models")
    session.add_all(forecasts)
    session.commit()

    logger.debug("Updating to latest")
    update_all_forecast_latest(
        session=session, forecasts=forecasts, update_national=update_national, update_gsp=update_gsp
    )
    session.commit()

    logger.debug("Saving to last seven days table")
    save_all_forecast_values_seven_days(session=session, forecasts=forecasts)
    session.commit()


def save_pv_system(session: Session, pv_system: PVSystem) -> PVSystemSQL:
    """
    Get model object from name and version

    :param session: database session
    :param pv_system: pv system sql object

    return: PVSystem object

    """

    # start main query
    query = session.query(PVSystemSQL)

    # filter on pv_system_id and provider
    query = query.filter(PVSystemSQL.pv_system_id == pv_system.pv_system_id)
    query = query.filter(PVSystemSQL.provider == pv_system.provider)

    # get all results
    pv_systems = query.all()

    if len(pv_systems) == 0:
        logger.debug(
            f"Model for provider {pv_system.provider} and pv_system_id {pv_system.pv_system_id }"
            f"does not exist so going to add it"
        )

        session.add(pv_system.to_orm())
        session.commit()

    else:
        logger.debug(
            f"Model for provider {pv_system.provider} and pv_system_id {pv_system.pv_system_id} "
            f"is already there so not going to add it"
        )
        pv_system = pv_systems[0]

    return pv_system


def save_all_forecast_values_seven_days(session: Session, forecasts: List[ForecastSQL]):
    """
    Save all the forecast values in the last seven days table

    :param session: database sessions
    :param forecasts: list of forecasts
    """

    # get all values together
    forecast_values_last_7_days = []
    for forecast in forecasts:
        for forecast_value in forecast.forecast_values:
            forecast_values_last_7_days.append(
                change_forecast_value_to_forecast_last_7_days(forecast_value)
            )

    # add them to the database
    add_forecast_last_7_days_and_remove_old_data(
        session=session, forecast_values=forecast_values_last_7_days
    )



""" Read user"""
import logging

from sqlalchemy.orm.session import Session

from nowcasting_datamodel.models.api import UserSQL

logger = logging.getLogger(__name__)


def get_user(session: Session, email: str) -> UserSQL:
    """
    Get metric object from database

    :param session: database session
    :param email: email of user

    return: One Metric SQl object

    """

    # start main query
    query = session.query(UserSQL)

    # filter on name
    query = query.filter(UserSQL.email == email)

    # get all results
    users = query.all()

    if len(users) == 0:
        logger.debug(f"User for name {email} does not exist so going to add it")

        user = UserSQL(email=email)

        session.add(user)
        session.commit()

    else:
        user = users[0]

    return user



"""Functions to make weights for blending"""
from datetime import datetime, timedelta, timezone
from typing import List, Optional

import numpy as np
import pandas as pd
import structlog

logger = structlog.stdlib.get_logger()

default_weights = [
    {
        "end_horizon_hour": 2,
        "end_weight": [1, 0],
    },
    {
        "start_horizon_hour": 2,
        "end_horizon_hour": 6,
        "start_weight": [1, 0],
        "end_weight": [0, 1],
    },
    {
        "start_horizon_hour": 6,
        "start_weight": [0, 1],
    },
]


def make_weights_df(
    model_names, weights, start_datetime=None, forecast_horizon_minutes: Optional[int] = None
):
    """Makes weights to half an hour and blocks

    A pd data frame like
    target_time weight cnn National_xg
    2020-01-01 00:00:00 1.0 0.0
    2020-01-01 00:30:00 0.5 0.5
    2020-01-01 01:00:00 0.0 1.0

    :param model_names: list of model names to use for blending
    :param start_datetime: start datetime to make weights from
    :param weights: dictionary of weights, see default for structure
    :param forecast_horizon_minutes: optional, if given then the weights are the same for the
    :return: pd dataframe of weights
    """
    if weights is None:
        weights = default_weights
    # get time now rounded up to 30 mins
    start_datetime_now = datetime.now(tz=timezone.utc).replace(second=0, microsecond=0)
    if start_datetime_now.minute >= 30:
        start_datetime_now += timedelta(hours=1)
        start_datetime_now = start_datetime_now.replace(minute=00)
    else:
        start_datetime_now = start_datetime_now.replace(minute=30)

    if start_datetime is None:
        start_datetime = start_datetime_now

    if forecast_horizon_minutes is not None:
        weights_for_forecast_horizon = get_weights_for_forecast_horizon(
            weights=weights, forecast_horizon_hours=int(forecast_horizon_minutes / 60)
        )
        logger.debug(
            f"weights for forecast_horizon {weights_for_forecast_horizon}"
            f" {forecast_horizon_minutes=}"
        )

    # make dataframe of 8 hours in 30 minutes chunks from now
    weights_all_df = []
    for i in range(len(weights)):
        weight = weights[i]
        end_horizon_hour, end_weight, start_horizon_hour, start_weight = extract_weight_variables(
            weight
        )

        if forecast_horizon_minutes is not None:
            end_weight = weights_for_forecast_horizon
            start_weight = weights_for_forecast_horizon

        # add the first weight to the dataframe, from  start_datetime to start_datetime_now.
        # This could be for the two days before now
        if i == 0:
            weights_df = pd.DataFrame(
                index=pd.date_range(start=start_datetime, end=start_datetime_now, freq="30min")
            )
            weights_df[model_names] = start_weight
            weights_all_df.append(weights_df)

        logger.debug(
            f"Making weights for {start_horizon_hour} to {end_horizon_hour} "
            f"hours with weights {start_weight} to {end_weight}"
        )

        start_datetime_one_weight = (
            start_datetime_now + timedelta(hours=start_horizon_hour) - timedelta(minutes=30)
        )
        if start_horizon_hour == 0:
            start_datetime_one_weight += timedelta(minutes=30)

        end_datetime = (
            start_datetime_now + timedelta(hours=end_horizon_hour) - timedelta(minutes=30)
        )
        if end_horizon_hour == 8:
            end_datetime += timedelta(hours=1)

        logger.debug(f"Making weights from {start_datetime_one_weight} to {end_datetime}")
        weights_df = pd.DataFrame(
            index=pd.date_range(start=start_datetime_one_weight, end=end_datetime, freq="30min")
        )

        # get rid of last timestamp
        weights_df = weights_df[:-1]

        assert len(model_names) == len(start_weight)
        assert len(model_names) == len(end_weight)
        for model_name, start_weight, end_weight in zip(model_names, start_weight, end_weight):
            weights_df[model_name] = start_weight + (end_weight - start_weight) * (
                weights_df.index - start_datetime_one_weight
            ) / (end_datetime - start_datetime_one_weight)

        weights_all_df.append(weights_df)

    weights_all_df = pd.concat(weights_all_df)
    weights_all_df = weights_all_df[~weights_all_df.index.duplicated(keep="first")]

    # only keep from now
    logger.debug(weights_all_df)
    logger.debug(start_datetime)
    weights_all_df = weights_all_df[weights_all_df.index >= start_datetime]

    return weights_all_df


def extract_weight_variables(weight):
    """
    Extra weight variables from dictionary

    :param weight: dictionary of weights
    :return:
    """
    if "start_horizon_hour" not in weight:
        start_horizon_hour = 0
    else:
        start_horizon_hour = weight["start_horizon_hour"]
    if "end_horizon_hour" not in weight:
        end_horizon_hour = 8
    else:
        end_horizon_hour = weight["end_horizon_hour"]
    if "start_weight" not in weight:
        start_weight = weight["end_weight"]
    else:
        start_weight = weight["start_weight"]
    if "end_weight" not in weight:
        end_weight = weight["start_weight"]
    else:
        end_weight = weight["end_weight"]
    return end_horizon_hour, end_weight, start_horizon_hour, start_weight


def get_weights_for_forecast_horizon(forecast_horizon_hours: int, weights) -> List[float]:
    """
    Get weights for a specific forecast horizon

    :param forecast_horizon_hours:
    :param weights: list of weights
    :return: for the one forecast horizon, the return weights for the different models
    """
    logger.debug(f"Getting weights for forecast horizon {forecast_horizon_hours} from {weights}")
    if weights is None:
        weights = default_weights

    for i in range(len(weights)):
        weight = weights[i]
        end_horizon_hour, end_weight, start_horizon_hour, start_weight = extract_weight_variables(
            weight
        )
        logger.debug(
            f"Checking {start_horizon_hour} to {end_horizon_hour} hours with weights "
            f"{start_weight} to {end_weight}"
        )

        if start_horizon_hour < forecast_horizon_hours < end_horizon_hour:
            logger.debug(
                f"Interpolating between {start_weight} and {end_weight} for forecast horizon "
                f"{forecast_horizon_hours}"
            )
            end_weight = np.array(end_weight)
            start_weight = np.array(start_weight)
            f = (start_horizon_hour - forecast_horizon_hours) / (
                start_horizon_hour - end_horizon_hour
            )
            logger.debug(f)
            return (f * end_weight + (1 - f) * start_weight).tolist()
        elif forecast_horizon_hours <= start_horizon_hour and (i == 0):
            logger.debug(
                f"Using end weight {end_weight} for forecast horizon "
                f"{forecast_horizon_hours} {end_horizon_hour=}"
            )
            return end_weight
        elif forecast_horizon_hours >= end_horizon_hour and (i == len(weights) - 1):
            logger.debug(
                f"Using start weight {start_weight} for forecast horizon "
                f"{forecast_horizon_hours} {start_horizon_hour=}"
            )
            return start_weight

    logger.debug(f"Did not find weights for forecast horizon {forecast_horizon_hours}")



""" Read pv functions """
import logging
from datetime import datetime, timezone
from typing import List, Optional, Union

from sqlalchemy import desc
from sqlalchemy.orm import Session, contains_eager, joinedload

from nowcasting_datamodel.models import GSPYieldSQL, LocationSQL

logger = logging.getLogger(__name__)


def get_latest_gsp_yield(
    session: Session,
    gsps: List[LocationSQL],
    append_to_gsps: bool = False,
    regime: str = "in-day",
    datetime_utc: Optional[datetime] = None,
) -> Union[List[GSPYieldSQL], List[LocationSQL]]:
    """
    Get the last gsp yield data

    :param session: database sessions
    :param gsps: list of gsps
    :param append_to_gsps: append gsp yield to pv systems, or return pv systems.
        If appended the yield is access by 'pv_system.last_gsp_yield'
    :param regime: What regime the data is in, either 'in-day' or 'day-after'
    :param datetime_utc: Optional to filter on datetime
    :return: either list of gsp yields, or pv systems
    """

    gsp_ids = [gsp.gsp_id for gsp in gsps]

    # start main query
    query = session.query(GSPYieldSQL)
    query = query.join(LocationSQL)
    query = query.where(
        LocationSQL.id == GSPYieldSQL.location_id,
    )

    # filter on regime
    query = query.where(GSPYieldSQL.regime == regime)

    # only select on results per pv system
    query = query.distinct(LocationSQL.gsp_id)

    # select only th epv systems we want
    query = query.where(LocationSQL.gsp_id.in_(gsp_ids))

    if datetime_utc is not None:
        # filter on datetime
        query = query.where(GSPYieldSQL.datetime_utc >= datetime_utc)

    # order by 'created_utc' desc, so we get the latest one
    query = query.order_by(
        LocationSQL.gsp_id, desc(GSPYieldSQL.datetime_utc), desc(GSPYieldSQL.created_utc)
    )

    # get all results
    gsp_yields: List[GSPYieldSQL] = query.all()

    logger.debug(f"Found {len(gsp_yields)}  latest gsp yields")

    # add utc timezone
    for gsp_yield in gsp_yields:
        gsp_yield.datetime_utc = gsp_yield.datetime_utc.replace(tzinfo=timezone.utc)

    if not append_to_gsps:
        return gsp_yields
    else:
        # get list of gsp with last pv yields
        gsp_systems_with_gsp_yields = []
        for gsp_yield in gsp_yields:
            gsp = gsp_yield.location
            gsp.last_gsp_yield = gsp_yield

            gsp_systems_with_gsp_yields.append(gsp)

        logger.debug(f"Found {len(gsp_systems_with_gsp_yields)} gsps with yields")

        # add pv systems that dont have any pv yields
        gsp_ids_with_gsp_yields = [gsp.gsp_id for gsp in gsp_systems_with_gsp_yields]

        gsp_systems_with_no_gsp_yields = []
        for gsp in gsps:
            if gsp.gsp_id not in gsp_ids_with_gsp_yields:
                gsp.last_gsp_yield = None

                gsp_systems_with_no_gsp_yields.append(gsp)

        logger.debug(f"Found {len(gsp_systems_with_no_gsp_yields)} gsps with no yields")

        all_gsp_systems = gsp_systems_with_gsp_yields + gsp_systems_with_no_gsp_yields

        return all_gsp_systems


def get_gsp_yield(
    session: Session,
    gsp_ids: List[int],
    start_datetime_utc: datetime,
    regime: Optional[str] = None,
    end_datetime_utc: Optional[datetime] = None,
    filter_nans: Optional[bool] = True,
) -> List[GSPYieldSQL]:
    """
    Get the gsp yield values.

    :param session: sqlalmcy sessions
    :param gsp_ids: list of gsp ids that we filter on
    :param start_datetime_utc: filter values on this start datetime
    :param regime: filter query on this regim. Can be "in-day" or "day-after"
    :param end_datetime_utc: optional end datetime filter
    :param filter_nans: optional filter out nans. Default is True
    :return: list of gsp yields
    """

    # start main query
    query = session.query(GSPYieldSQL)
    query = query.join(LocationSQL)
    query = query.options(joinedload(GSPYieldSQL.location))
    query = query.where(
        LocationSQL.id == GSPYieldSQL.location_id,
    )

    # filter on regime
    if regime is not None:
        query = query.where(GSPYieldSQL.regime == regime)

    # filter on datetime
    query = query.where(GSPYieldSQL.datetime_utc >= start_datetime_utc)
    if end_datetime_utc is not None:
        query = query.where(GSPYieldSQL.datetime_utc <= end_datetime_utc)

    # don't get any nans. (Note nan+1 > nan = False)
    if filter_nans:
        query = query.where(GSPYieldSQL.solar_generation_kw + 1 > GSPYieldSQL.solar_generation_kw)

    # only select on results per gsp
    query = query.distinct(*[LocationSQL.gsp_id, GSPYieldSQL.datetime_utc])

    # select only the gsp systems we want
    query = query.where(LocationSQL.gsp_id.in_(gsp_ids))

    # order by 'created_utc' desc, so we get the latest one
    query = query.order_by(
        LocationSQL.gsp_id, desc(GSPYieldSQL.datetime_utc), desc(GSPYieldSQL.created_utc)
    )

    # get all results
    gsp_yields: List[GSPYieldSQL] = query.all()

    for gsp_yield in gsp_yields:
        gsp_yield.datetime_utc = gsp_yield.datetime_utc.replace(tzinfo=timezone.utc)

    return gsp_yields


def get_gsp_yield_by_location(
    session: Session,
    gsp_ids: List[int],
    start_datetime_utc: datetime,
    regime: Optional[str] = None,
    end_datetime_utc: Optional[datetime] = None,
    filter_nans: Optional[bool] = True,
) -> List[LocationSQL]:
    """
    Get the locations with gsp yield values.

    :param session: sqlalmcy sessions
    :param gsp_ids: list of gsp ids that we filter on
    :param start_datetime_utc: filter values on this start datetime
    :param regime: filter query on this regim. Can be "in-day" or "day-after"
    :param end_datetime_utc: optional end datetime filter
    :param filter_nans: optional filter out nans. Default is True
    :return: list of locations with gsp yields
    """

    # start main query
    query = session.query(LocationSQL)
    query = query.join(GSPYieldSQL)
    query = query.where(LocationSQL.id == GSPYieldSQL.location_id)

    # filter on regime
    if regime is not None:
        query = query.where(GSPYieldSQL.regime == regime)

    # filter on datetime
    query = query.where(GSPYieldSQL.datetime_utc >= start_datetime_utc)
    if end_datetime_utc is not None:
        query = query.where(GSPYieldSQL.datetime_utc <= end_datetime_utc)

    # don't get any nans. (Note nan+1 > nan = False)
    if filter_nans:
        query = query.where(GSPYieldSQL.solar_generation_kw + 1 > GSPYieldSQL.solar_generation_kw)

    # only select on results per gsp
    query = query.distinct(*[LocationSQL.gsp_id, GSPYieldSQL.datetime_utc])

    # select only the gsp systems we want
    query = query.where(LocationSQL.gsp_id.in_(gsp_ids))

    # load only once
    query = query.options(contains_eager(LocationSQL.gsp_yields))

    # order by 'created_utc' desc, so we get the latest one
    query = query.order_by(
        LocationSQL.gsp_id, desc(GSPYieldSQL.datetime_utc), desc(GSPYieldSQL.created_utc)
    )

    # get all results
    locations: List[LocationSQL] = query.all()

    for location in locations:
        for gsp_yield in location.gsp_yields:
            gsp_yield.datetime_utc = gsp_yield.datetime_utc.replace(tzinfo=timezone.utc)

    return locations
